from flask import jsonify, request

def format_errors(*errors):
    return {
        '_errors': errors
    }

def format_error(error):
    return format_errors(error)


def output_json(data, code, headers=None):
    """
    
    .. code-block:: python
        
        api.representations['application/json'] = output_json
    
    Generates better formatted responses for RESTFul APIs. 

    If the restful resource responds with a string, with a non 200 error,
    the response will look like

    .. code-block:: json

        {
            "_errors": ["String the user responded with"]
        }

    Likewise, a string return value with a 200 response will look like:


    .. code-block:: json

        {
            "message": "String the user returned with."
        }

    
    If a Non-200 response occured, and flask-restful added it's own error 
    message in the "message" field of the response data, this data is moved into
    "_errors":


    .. code-block:: json

        {
            "_errors": ["You don't have the permission to access the requested resource..."]
        }


    All data is returned using flask's jsonify method. This means you can 
    use simplejson to return decimal objects from your flask restful resources.
    """

    if type(data) is str:
        # Handle view returning a string.
        message = data
        if code != 200:
            data = format_errors(message)
        else:
            data = {'message': message}

    elif code != 200 and type(data) is dict and 'message' in data:
        # Flask-Restful returns non-200 error messages to the user.
        # Let's show them.
        data = format_errors(data.get('message'))

    resp = jsonify(data)

    if code:
        resp.status_code = code

    resp.headers.extend(headers)
    return resp



class ExpectedJSONException(Exception):
    """
    Thrown when JSON was expected in a flask request but was not
    provided
    """
    pass

    @classmethod
    def handle(cls, exc):
        """
        A handler for this type of exception.

        Usage:

        .. code-block:: python
     
            app.errorhandler(ExpectedJSONException)(ExpectedJSONException.handle)


        """
        return jsonify(
            format_error('Resource expected a JSON payload to be provided.')
        ), 400


def get_and_expect_json():
    """
    Returns the ``flask.request.get_json()`, however if no JSON data was decoded,
    will raise a :class:`ExpectedJSONException`
    """

    data = request.get_json()
    if data is None:
        raise ExpectedJSONException()
    
    return data




from celery import Celery
from twopi_flask_utils.deployment_release import get_release

def create_celery(name, config_obj, inject_version=True, **kwargs):
    """Creates a celery app.
    
    :param config_obj: The configuration object to initiaze with
    :param inject_version: bool: Whether or not to inject the application's
                                 version number. Attempts to get version number
                                using 
                                :func:`twopi_flask_utils.deployment_release.get_release`
    :param kwargs: Other arguments to pass to the ``Celery`` instantiation.
    :returns: An initialized celery application.
    """

    celery = Celery(name, broker=config_obj.CELERY_BROKER_URL, **kwargs)
    celery.config_from_object(config_obj)
    if inject_version:
        celery.version = get_release()

    return celery


def create_db_session(celery):
    """
    Creates an SQLA Database scoped session. Requires SQLAlchemy.

    Uses ``SQLALCHEMY_DATABASE_URI`` and ``SQLALCHEMY_POOL_RECYCLE`` to create an 
    apropriate engine. 

    If you are using MySQL and ``SQLALCHEMY_POOL_RECYCLE`` is not specified, 
    you'll have a bad time - this is required, as MySQL kills old sessions.

    :param celery: The celery application to create the app on
    :returns: A SQLAlchemy scoped_session instance.

    """

    from sqlalchemy import create_engine
    from sqlalchemy.orm import scoped_session
    from sqlalchemy.orm import sessionmaker

    kwargs = {}
    if celery.conf.get('SQLALCHEMY_POOL_RECYCLE'):
        kwargs['pool_recycle'] = celery.conf.get('SQLALCHEMY_POOL_RECYCLE')

    engine = create_engine(
        celery.conf.get('SQLALCHEMY_DATABASE_URI'), 
        **kwargs
    )
    return scoped_session(sessionmaker(
        autocommit=False, autoflush=False, bind=engine))



from raven import Client 
from raven.contrib.celery import register_signal, register_logger_signal
from raven.contrib.flask import Sentry


def create_client(conf, app_version='__UNKNOWN__'):
    """Creates a sentry client.

    :param app_version: (string): App version sent to sentry for making events more rich
    :param conf['SENTRY_DSN']: (string, required): DSN of sentry server
    :param conf['SENTRY_SITE']: (string): The site description of the deployment.
    :returns: An initialized ``raven.Client`` instance.
    """
    client = Client(
        conf['SENTRY_DSN'],
        site=conf.get('SENTRY_SITE'),
        release=app_version,
    )
    return client

def inject_sentry(app):
    """Injects sentry into a Flask Application
    
    Will only inject if ``SENTRY_DSN`` is specified. ``SENTRY_SITE`` and 
    ``app.version`` are used to provide extra context to sentry events.

    :param app: (Flask Instance): A flask application to attach raven to.
    """

    if app.config.get('SENTRY_DSN'):
        client = create_client(app.config, app.version)
        Sentry(app, client=client)


def celery_inject_sentry(celery):
    """Inject Sentry into a celery app. Requires ``raven``.
    
    If ``SENTRY_DSN`` is specified in config, a sentry client is created and 
    attached. Additionally uses ``celery.version`` and ``SENTRY_SITE`` to 
    provide extra context to sentry events.
    
    :param celery: The celery instance to attach raven to.

    """
    if celery.conf.get('SENTRY_DSN'):
        client = create_client(celery.conf, 
                               app_version=getattr(celery, 'version', 'UNKNOWN'))

        register_logger_signal(client)
        register_signal(client)



from flask import g, request, jsonify
from functools import wraps
from twopi_flask_utils.restful import format_error
import re

bearer_re = re.compile(r'Bearer (.+)')

def parse_auth_header(token_cls, auth_header=True, query_string=True, secret=None):
    """
    A decorator to extract a token from either the ``Authorization`` header OR
    the query string parameter ``?token=``.

    :param token_cls: An instance of :class:`ShortlivedTokenMixin` OR a class which
                      implements the classmethod ``load(token_string, secret)``. 
                      An instance will be available on `g.token`.
    :param auth_header:  (optional, ``bool``) Extract the token from the auth 
                          header (Default: ``True``)
    :param query_string: (optional, ``bool``) Extract the token from the query 
                          string (Default: ``True``)
    :param secret: (optional) A secret to pass to ``token_cls.load(raw, secret)``

    Any wrapped function will be able to access both ``g.token`` and ``g.raw_token``
    to read the ``token_cls`` instance and raw token string respectively.

    Authorization header expects tokens in the format of ``Bearer <token string>``

    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            g.token = None
            g.raw_token = None
            raw_token = None

            if query_string and 'token' in request.args:
                # Try fetch the token from the QS
                raw_token = request.args['token']

            if auth_header and 'Authorization' in request.headers:
                token_res = bearer_re.match(request.headers['Authorization'])
                if token_res is not None:
                    raw_token = token_res.group(1)

            if raw_token is not None:
                token = token_cls.load(raw_token, secret)
                if token is None:
                    return jsonify(format_error("The provided token was invalid.")), 401

                g.token = token
                g.raw_token = raw_token

            return f(*args, **kwargs)

        return wrapped
    return wrapper


def auth_required():
    """
    Force authentication on an endpoint. Checks if ``g.token`` is not None, 
    and returns a ``401`` if it is.

    Example:

    .. code::

        @app.route('/auth-required')
        @auth_required()
        def my_endpoint():
            return "Hello World"

    """
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            if g.token is None:
                return jsonify(
                    format_error("A valid token is required to access this resource")), 401

            return f(*args, **kwargs)
        
        return wrapped
    return wrapper



import pytz
import datetime
import jwt
from marshmallow import fields, Schema, post_dump
from flask import current_app
import logging

log = logging.getLogger(__name__)

EPOCH_DT = datetime.datetime(1970, 1, 1, tzinfo=pytz.UTC)

class UnixTimestamp(fields.Field):
    def _serialize(self, value, attr, obj):
        if value is None:
            return None

        return int((value - EPOCH_DT).total_seconds())

    def _deserialize(self, value, attr, data):
        if value is None:
            return None

        return datetime.datetime.fromtimestamp(value, pytz.UTC)


class ShortlivedTokenMixin(object):
    """
    A base class for implementing a short-lived token using JWTs

    :param expiry: ``datetime``: The expiry of this token
    :param issuer: ``string``
    :param subject: ``string``
    :param audience: ``string``
    :param not_before: ``datetime``: A datetime of when the token becomes 
                                     valid for use
    :param issued_at: ``datetime``: When the token was issued. This value is 
                                    overwritten during ``dump()``
    """
    class TokenSchema(Schema):
        """
        The schema to use to serialize/de-serialize JWT's with.
        """
        SKIPPABLE = ['exp', 'iss', 'sub', 'aud', 'nbf', 'iat']

        exp = UnixTimestamp(attribute='expiry')
        iss = fields.String(attribute='issuer')
        sub = fields.String(attribute='subject')
        aud = fields.String(attribute='audience')
        nbf = UnixTimestamp(attribute='not_before')
        iat = UnixTimestamp(attribute='issued_at')

        @post_dump
        def remove_skippable(self, data):
            return {key: value for key, value in data.items() 
                    if key not in self.SKIPPABLE or value is not None}

    def __init__(self, expiry=None, issuer=None, subject=None, audience=None, 
                 not_before=None, issued_at=None):
        self.expiry = expiry
        self.issuer = issuer
        self.subject = subject
        self.audience = audience
        self.not_before = not_before
        self.issued_at = issued_at

    @classmethod
    def generate(Cls, refresh_token):
        log.warning("{}.generate() is deprecated. Please use "
                    "{}.from_refresh_token()".format(Cls.__name__, Cls.__name__))
        return Cls.from_refresh_token(refresh_token)

    @classmethod
    def from_refresh_token(Cls, refresh_token):
        """
        Given a refresh token, return an instance of ShortLivedTokenMixin 
        configured using information from ``refresh_token``.

        :param refresh_token: A refresh token instance.
        :returns: A new ShortLivedToken instance.

        .. warning::
            
            You must implement this method



        """
        raise NotImplemented("You must implement from_refresh_token. Invoke {}() "
                             "and instantiate it using information from the "
                             "refresh token".format(Cls.__name__))

    @classmethod
    def load(Cls, token_string, secret=None, issuer=None, audience=None):
        """
        Load from a JWT (``token_string``)

        :param token_string: The raw string to load from
        :param secret: The secret that the JWT was signed with to check validity. 
                       If this is omitted, the secret will be sourced 
                       from ``current_app.config['SECRET_KEY']``
        :param issuer: The issuer the JWT decode should expect
        :param audience: The audience the JWT decode should expect
        :returns: A de-serialized ShortLivedToken instance.
        """

        if secret is None:
            secret = current_app.config['SECRET_KEY']

        try:
            payload = jwt.decode(token_string, secret, issuer=issuer, audience=audience)
        except (jwt.exceptions.InvalidTokenError) as e:
            log.info("The provided token has expired or was malformed. {}".format(e))
            return None

        deserialized, errs = Cls.TokenSchema().load(payload)
        if errs:
            # Malformed token?
            log.info("Malformed token was provided, error during de-serialisation.")
            return None
        try:
            return Cls(**deserialized)
        except TypeError as e:
            log.info("Malformed token was provided, error during instantiation.")
            return None

    def dump(self, secret=None):
        """
        Dump the token into a stringified JWT.

        :param secret: The secret to sign the JWT with. If this is omitted, the 
                       secret will be sourced from ``current_app.config['SECRET_KEY']``

        :returns: The stringified JWT.
        """

        self.issued_at = datetime.datetime.now(pytz.UTC)
        payload, err = self.TokenSchema().dump(self)

        if secret is None:
            secret = current_app.config['SECRET_KEY']

        return jwt.encode(payload, secret).decode('UTF-8')


    def __repr__(self):
        return "<ShortlivedToken expiry={}>".format(self.expiry)



def get_release():
    """
    Opens a file ``version.txt`` and returns it's stripped contents.
    
    :returns: The stripped file contents
    """
    try:
        with open('version.txt') as fh:
            return fh.read().strip()
    except Exception:
        return '__UNKNOWN__'



from webargs.flaskparser import FlaskParser
from webargs.core import ValidationError
from flask import jsonify


class BetterFlaskParser(FlaskParser):
    """
    A Flask-Restful compatible parser for WebArgs.
    """

    def handle_error(self, error):
        """
        Don't raise a ``HTTPException`` via ``abort``. Instead we will throw the 
        ``ValidationError`` and handle it with our flask Exception handler.

        This allows a common code path for both Flask-Restful AND standard 
        Flask Views.
        """
        raise error


parser = BetterFlaskParser()
use_args = parser.use_args
use_kwargs = parser.use_kwargs

def handle_validation_error(exc):
    """
    When using :class:`BetterFlaskParser`, if an exception occurs, it will throw the 
    original ``ValidationError``. This circumvents the capture inside Flask-Restful
    (if it is being used at all). 

    Instead of capturing all 422 ``HTTPExceptions``, you register this error 
    handler with ``ValidationError``:

    .. code-block:: python
    
        app.errorhandler(ValidationError)(handle_validation_error)


    This function will produce a jsonified response with the field errors from
    the ValidationError.

    .. warning::

        This handler is incompatible with the standard ``FlaskParser``, since
        it throws ``HTTPExceptions`` (via ``abort``). Registering this handler with 
        ``errorhandler(422)`` will not work.

    """

    assert(type(exc) == ValidationError)
    return jsonify(exc.messages), exc.status_code


__all__ = ['BetterFlaskParser', 'parser', 'use_args', 'use_kwargs',
           'handle_validation_error']




from setuptools import setup
import os

extras = {
    'ldap': [],
    'restful': ['flask-restful'],
    'celery': ['celery'],
    'sentry': ['raven[flask]'],
    'pagination': ['webargs', 'marshmallow'],
    'webargs': ['webargs'],
    'token_auth': ['marshmallow']
}

packages = [
    'twopi_flask_utils',
    'twopi_flask_utils.celery',
    'twopi_flask_utils.config',
    'twopi_flask_utils.deployment_release',
    'twopi_flask_utils.pagination',
    'twopi_flask_utils.restful',
    'twopi_flask_utils.sentry',
    'twopi_flask_utils.testing',
    'twopi_flask_utils.token_auth',
    'twopi_flask_utils.webargs',
]

tests_require = []
for x in extras.values():
    tests_require.extend(x)


def get_version():
    version_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'VERSION')
    v = open(version_path).read()
    if type(v) == str:
        return v.strip()
    return v.decode('UTF-8').strip()


readme_path = os.path.join(os.path.dirname(
  os.path.abspath(__file__)),
  'README.rst',
)
long_description = open(readme_path).read()

try:
    version = get_version()
except Exception as e:
    version = '0.0.0-dev'


setup(
    name='twopi-flask-utils',
    version=version,
    packages=packages,
    package_data={'': ['LICENSE', 'README.rst']},
    zip_safe=False,
    include_package_data=True,
    extras_require=extras,
    tests_require=tests_require,
    test_suite='tests',
    url='https://github.com/TwoPiCode/twopi-flask-utils',
    author='Nick Whyte',
    author_email='nick@twopicode.com',
    description=('A set of utilities to make working with flask web '
                 'applications easier.'),
    long_description=long_description
)



import json

class AppReqTestHelper(object):
    """
    Adds convenience request methods on the testcase object.

    Assumes a flask app client is defined on ``self.client``
    """
    def _req(self, meth, *args, **kwargs):
        if kwargs.get('content_type') is None and meth != 'get':
            kwargs['content_type'] = 'application/json'

        if kwargs.get('json') is not None:
            kwargs['data'] = json.dumps(kwargs.get('json'))
            del kwargs['json']

        # Take provided headers or fall back.
        headers = kwargs.pop('headers', getattr(self, 'client_headers', {}))

        func = getattr(self.client, meth)
        rv = func(*args, headers=headers, **kwargs)

        def get_json():
            return json.loads(rv.data.decode('UTF-8'))

        rv.get_json = get_json
        return rv

    def post(self, *args, **kwargs):
        """Perform a post request to the application."""
        return self._req('post', *args, **kwargs)

    def get(self, *args, **kwargs):
        """Perform a get request to the application."""
        return self._req('get', *args, **kwargs)

    def put(self, *args, **kwargs):
        """Perform a put request to the application."""
        return self._req('put', *args, **kwargs)

    def delete(self, *args, **kwargs):
        """Perform a delete request to the application."""
        return self._req('delete', *args, **kwargs)

    def patch(self, *args, **kwargs):
        """Perform a patch request to the application."""
        return self._req('patch', *args, **kwargs)


class PrivilegeTestHelper(AppReqTestHelper):
    """
    Adds a helper to test endpoint privileges in an application.
    """

    def do_test_privileges(self, endpoint, data, object_id, expected_codes):
        """
        Test privileges on a specific endpoint.

        :param endpoint: The endpoint to test. e.g. `/api/v1/classes`
        :param data: The data to use when performing a `PUT`/`POST` (dict)
        :param object_id: The id of the singular object to test permissions on `PUT`/`DELETE`/`GET`
        :param expected_codes: The expected response codes when performing each
                               endpoint request. E.g.
                                .. code::

                                    {
                                        'plural-get': 200,
                                        'get': 200,
                                        'delete': 403,
                                        'post': 200,
                                        'put': 200,
                                    }

        """
        for meth, expected_code in expected_codes:
            assert meth in ['plural-get', 'get', 'delete', 'put', 'post']

            kwargs = {}
            endp = endpoint
            _meth = meth
            if meth in ['put', 'post']:
                kwargs['json'] = data

            if meth in ['put', 'delete', 'get']:
                endp = endpoint + '/{}'.format(object_id)

            if meth == 'plural-get':
                _meth = 'get'

            rv = self._req(_meth, endp, **kwargs)
            self.assertEqual(rv.status_code, expected_code, 
                             "Expected {} for method {} but got {}. {}".format(
                                expected_code, meth, rv.status_code, rv.get_json()))


class CRUDTestHelper(AppReqTestHelper):
    """
    A helper to test generic CRUD operations on an endpoint.
    """
    def do_crud_test(self, endpoint, data_1=None, data_2=None, key='id', 
                     check_keys=[], keys_from_prev=[], create=True, delete=True, 
                     update=True, read=True, initial_count=0):
        """
        Begins the CRUD test. 

        :param endpoint: ``string``: The endpoint to test
        :param data1: ``dict``: Data to create the initial entity with (POST)
        :param data2: ``dict``: Data to update the entity with (PUT)
        :param key: ``string``: The key field in the response returned when performing
                            a create.
        :param check_keys: ``list``: A list of keys to compare ``data_1`` and 
                                 ``data_2`` to returned API responses. (To 
                                 ensure expected response data)
        :param keys_from_prev: ``list``: A list of keys to check that they persisted 
                                     after a create/update.
        :param create: ``bool``: Should create a new object and test it's existence
        :param delete: ``bool``: Should delete the newly created object and test 
                             that it has been deleted.
        :param update: ``bool``: Should performs PUT (update)
        :param read: ``bool``: Should perform a plural read
        :param initial_count: ``int``: The initial number of entities in the endpoint's 
                                   dataset
        """

        if read:
            # Plural read on initial set.
            rv = self.get(endpoint)
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(len(rv.get_json()), initial_count)

        if create:
            # Create
            rv = self.post(endpoint, json=data_1)
            self.assertEqual(rv.status_code, 200)
            self.assertEqualDicts(check_keys, rv.get_json(), data_1)
            key_id = rv.get_json()[key]

        if read:
            # Plural read
            rv = self.get(endpoint)
            self.assertEqual(rv.status_code, 200)
            if create:
                self.assertEqual(len(rv.get_json()), initial_count + 1)
                # Ensure that at least one of the items is equal to the one we made
                for item in rv.get_json():
                    if self.equalDicts(check_keys, item, data_1):
                        break
                else: # nobreak
                    self.fail("Could not find the object that was created in the response.")

            else:
                # No object should have been created
                self.assertEqual(len(rv.get_json()), initial_count)

        if read and create:
            # Singular Read
            rv = self.get(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 200)
            prev_data = rv.get_json() # Keep this data so we can use it after update.
            self.assertEqualDicts(check_keys, prev_data, data_1)

        if update and create:
            # Singular Update
            rv = self.put(endpoint + '/' + str(key_id), json=data_2)
            self.assertEqual(rv.status_code, 200)
            self.assertEqualDicts(check_keys, rv.get_json(), data_2)
            self.assertEqualDicts(keys_from_prev, rv.get_json(), prev_data)
        
        if read and create:
            # Singular Read to confirm persisted.
            rv = self.get(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 200)
            self.assertEqualDicts(check_keys, rv.get_json(), data_2)
            self.assertEqualDicts(keys_from_prev, rv.get_json(), prev_data)

        if delete and create:
            # Singular Deletion
            rv = self.delete(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 200)

        if read and create:
            # Singular read on object which doesn't exist.
            rv = self.get(endpoint + '/' + str(key_id))
            self.assertEqual(rv.status_code, 404)

        if read:
            # Plural read on empty set
            rv = self.get(endpoint)
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(len(rv.get_json()), initial_count)

    def filteredDicts(self, keys, *dicts):
        ret = []
        for d in dicts:    
            d_filtered = dict((k, v) for k, v in d.items()
                              if k in keys)
            ret.append(d_filtered)
        
        return ret

    def assertEqualDicts(self, keys, d1, d2):
        self.assertEqual(*self.filteredDicts(keys, d1, d2))

    def equalDicts(self, keys, d1, d2):
        d1, d2 = self.filteredDicts(keys, d1, d2)
        return d1 == d2

